﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography;
using System.Text;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using Microsoft.ParallelArrays;

using BPA = Microsoft.ParallelArrays.BoolParallelArray;
using DPA = Microsoft.ParallelArrays.DoubleParallelArray;
using F4PA = Microsoft.ParallelArrays.Float4ParallelArray;
using FPA = Microsoft.ParallelArrays.FloatParallelArray;
using IPA = Microsoft.ParallelArrays.IntParallelArray;
using PA = Microsoft.ParallelArrays.ParallelArrays;

namespace ImageLab.Filter
{
    public static class FilterCore
    {
        public static DX9Target DX9 = new DX9Target();
       // public static X64MulticoreTarget X64 = new X64MulticoreTarget();

        public static RNGCryptoServiceProvider RNG = new RNGCryptoServiceProvider();

        private static WriteableBitmap input, output;

        public static WriteableBitmap Output
        {
            get { return output; }
        }
        public static WriteableBitmap Input
        {
            set { input = value; }
        }

        public static bool CalculateHistogram = false;
        public static Float4[] Histogram;

        public delegate F4PA FilterDelegate(F4PA input, params object[] parameters);
        public static void Process(FilterDelegate filter, params object[] parameters)
        {
            int _Width, _Height; // Sub image size
            var _Buffers = PartitionImage(1000000, out _Width, out  _Height); // Partitioning image into subfields

            var pr = Profile.Start("Filtering operation");
            Histogram = new Float4[Analysis.HistogramBins];// Emptying Histogram
            var ShapeOut = new int[2];
            for (int i = 0; i < _Buffers.Length; i++)
            {
                var _PImageIn = RGBA128ToF4PA(_Buffers[i], _Width, _Height);
                var _PImageOut = filter(_PImageIn, parameters);// Performing filtering operation in delegate
                _PImageOut = LimitF4PA(_PImageOut); // Limiting function
                if (CalculateHistogram)
                {
                    var _TempHistogram = Analysis.Histogram(_PImageOut);
                    for (int j = 0; j < _TempHistogram.Length; j++)
                        Histogram[j] += _TempHistogram[j];
                }
                ShapeOut = _PImageOut.Shape;
                _Buffers[i] = F4PAToRGBA128(_PImageOut, ShapeOut[1], ShapeOut[0]);
            }
            pr.Stop();

            BuildImage(_Buffers, ShapeOut[1], ShapeOut[0]);

            Console.WriteLine("Filtering done");
        }
        private static void BuildImage(byte[][] Buffers, int _Width, int _Height)
        {
            var _Scale = (int)Math.Sqrt(Buffers.Length);

            output = new WriteableBitmap(
                _Width * _Scale,
                _Height * _Scale,
                input.DpiX,
                input.DpiY,
                input.Format,
                null);

            output.Lock();
            var ctr = 0;
            var _Side = Math.Sqrt(Buffers.Length);
            var _Stride = _Width * output.Format.BitsPerPixel / 8;
            for (int i = 0; i < _Side; i++)
            {
                for (int j = 0; j < _Side; j++)
                {
                    var _Rectangle = new Int32Rect(i * _Width, j * _Height, _Width, _Height);
                    output.WritePixels(_Rectangle, Buffers[ctr], _Stride, 0);
                    ctr++;
                }
            }
            output.AddDirtyRect(new Int32Rect(0, 0, output.PixelWidth, output.PixelHeight));
            output.Unlock();
        }
        private static byte[][] PartitionImage(int MaxPixels, out int _Width, out int _Height)
        {
            _Width = input.PixelWidth;
            _Height = input.PixelHeight;

            var _SubImages = 1;
            while (_Width * _Height > MaxPixels)
            {
                _Width = _Width / 2;
                _Height = _Height / 2;
                _SubImages *= 4;
            }

            var _Pixels = _Width * _Height;
            var _Buffers = new byte[_SubImages][];
            for (int i = 0; i < _Buffers.Length; i++)
                _Buffers[i] = new byte[_Pixels * input.Format.BitsPerPixel / 8];

            var ctr = 0;
            var _Side = Math.Sqrt(_SubImages);
            var _Stride = _Width * input.Format.BitsPerPixel / 8;
            for (int i = 0; i < _Side; i++)
            {
                for (int j = 0; j < _Side; j++)
                {
                    var _Rectangle = new Int32Rect(i * _Width, j * _Height, _Width, _Height);
                    input.CopyPixels(_Rectangle, _Buffers[ctr], _Stride, 0);
                    ctr++;
                }
            }

            return _Buffers;
        }
        
        public static F4PA PowerF4PA(F4PA input, FPA exp)
        {
            var r = PA.FromFloat4ParallelArray(input, 1);
            var g = PA.FromFloat4ParallelArray(input, 2);
            var b = PA.FromFloat4ParallelArray(input, 3);
            var a = PA.FromFloat4ParallelArray(input, 4);

            r = PA.Pow(r, exp);
            g = PA.Pow(g, exp);
            b = PA.Pow(b, exp);
            a = PA.Pow(a, exp);

            return PA.ToFloat4ParallelArray(r, g, b, a);
        }
        public static F4PA LimitF4PA(F4PA input)
        {
            var zero = new F4PA(new Float4(0), input.Shape);
            var one = new F4PA(new Float4(1), input.Shape);

            input = PA.Select(input, input, zero); // limiting to [0,1]
            input = PA.Select(one - input, input, one);

            input *= new Float4(1, 1, 1, 0);// Resetting alpha value
            input += new Float4(0, 0, 0, 1);

            return input;
        }

        public static Float4[,] RGBA128ToFloat4(byte[] bytes, int width, int height)
        {
            var rgba = RGBA128ToF4PA(bytes, width, height);
            // returning an 2D-array of Float4 structures
            return FilterCore.DX9.ToArray2D(rgba);
        }

        public static F4PA RGBA128ToF4PA(byte[] bytes, int width, int height)
        {
            // Reserving temporary memory for storing the floats
            var RGBA128_float = new float[height, 4 * width];
            // Copying the bytes into the temporary memory
            Buffer.BlockCopy(bytes, 0, RGBA128_float, 0, bytes.Length);
            // Creating a parallel array for gpgpu processing
            var pa = new FPA(RGBA128_float);
            // Specifying how to partition the image
            var ssr = new SectionSpecifier(0, -1, 4);// every 4th element, offset 0
            var ssg = new SectionSpecifier(1, -1, 4);// every 4th element, offset 1
            var ssb = new SectionSpecifier(2, -1, 4);// every 4th element, offset 2
            var ssa = new SectionSpecifier(3, -1, 4);// every 4th element, offset 3
            var ssother = new SectionSpecifier(0, -1, 1);// every element
            // Partitioning the image
            var r = PA.Section(pa, ssother, ssr);
            var g = PA.Section(pa, ssother, ssg);
            var b = PA.Section(pa, ssother, ssb);
            var a = PA.Section(pa, ssother, ssa);
            // Converting to color data
            return PA.ToFloat4ParallelArray(r, g, b, a);
        }
        public static byte[] F4PAToRGBA128(F4PA input, int width, int height)
        {
            var image_float = new float[height, 4 * width];
            var image_filtered = FilterCore.DX9.ToArray2D(input);

            for (int i = 0; i < height; i++)
            {
                for (int j = 0; j < width; j++)
                {
                    image_float[i, j * 4] = image_filtered[i, j].f1;
                    image_float[i, j * 4 + 1] = image_filtered[i, j].f2;
                    image_float[i, j * 4 + 2] = image_filtered[i, j].f3;
                    image_float[i, j * 4 + 3] = image_filtered[i, j].f4;
                }
            }

            var buffer = new byte[height * width * 16];
            Buffer.BlockCopy(image_float, 0, buffer, 0, buffer.Length);

            return buffer;
        }
        /*
        public static ImageData CreateImageData(F4PA filtered, PixelFormat format)
        {
            var shape = filtered.Shape;
            var bytes = F4PAToRGBA128(filtered, shape[1], shape[0]);

            var output = new ImageData();
            output.Width = shape[1];
            output.Height = shape[0];
            output.PixelFormat = format;
            output.Bytes = bytes;
            return output;
        }
        */
        public static void Sort(F4PA[] input)
        {
            for (int i = 0; i < input.Length; i++)
            {
                for (int j = input.Length - 1; j >= i + 1; j--)
                {
                    var min = PA.Min(input[i], input[j]);
                    var max = PA.Max(input[i], input[j]);
                    input[i] = min;
                    input[j] = max;
                }
            }
        }
        public static void Sort(F4PA[] sortable, F4PA[] extra)
        {
            for (int i = 0; i < sortable.Length; i++)
            {
                for (int j = sortable.Length - 1; j >= i + 1; j--)
                {
                    var diff = sortable[i] - sortable[j];
                    var min1 = PA.Select(diff, sortable[j], sortable[i]);
                    var min2 = PA.Select(diff, extra[j], extra[i]);
                    var max1 = PA.Select(diff, sortable[i], sortable[j]);
                    var max2 = PA.Select(diff, extra[i], extra[j]);
                    sortable[i] = min1;
                    sortable[j] = max1;
                    extra[i] = min2;
                    extra[j] = max2;
                }
            }
        }
        public static F4PA Divide(F4PA input1, F4PA input2)
        {
            var r = PA.FromFloat4ParallelArray(input1, 1) / PA.FromFloat4ParallelArray(input2, 1);
            var g = PA.FromFloat4ParallelArray(input1, 2) / PA.FromFloat4ParallelArray(input2, 2);
            var b = PA.FromFloat4ParallelArray(input1, 3) / PA.FromFloat4ParallelArray(input2, 3);
            var a = PA.FromFloat4ParallelArray(input1, 4) / PA.FromFloat4ParallelArray(input2, 4);

            return PA.ToFloat4ParallelArray(r, g, b, a);
        }
    }
}
