﻿
using Microsoft.ParallelArrays;
using F4 = Microsoft.ParallelArrays.Float4;
using F4PA = Microsoft.ParallelArrays.Float4ParallelArray;
using PA = Microsoft.ParallelArrays.ParallelArrays;

namespace ImageLab.Filter
{
    public static class MeanFilter
    {
        public static F4PA Mean(F4PA input, params object[] parameters)
        {
            var radius = (int)parameters[0];

            var sum1 = new F4PA(new F4(0f), input.Shape);
            var sum2 = new F4PA(new F4(0f), input.Shape);

            var n_samples = (2 * radius + 1) * (2 * radius + 1);

            for (int i = -radius; i <= radius; i++)
            {
                sum1 += PA.Shift(input, new int[] { i, 0 });
            }
            for (int i = -radius; i <= radius; i++)
            {
                sum2 += PA.Shift(sum1, new int[] { 0, i });
            }
            return sum2 / new F4(n_samples);
        }

        public static F4PA WinsorizedMean(F4PA input, params object[] parameters)
        {
            var radius = (int)parameters[0];
            var alfa = (float)parameters[1];

            var n_samples = (2 * radius + 1) * (2 * radius + 1);
            var n_min = (int)(n_samples * alfa);
            var n_max = (int)(n_samples - n_min);
            var n_used = n_max - n_min;

            var images = new F4PA[n_samples];

            var ctr = 0;
            for (int i = -radius; i <= radius; i++)
                for (int j = -radius; j <= radius; j++)
                    images[ctr++] = PA.Shift(input, new int[] { i, j });

            FilterCore.Sort(images);

            var mean = new F4PA(new F4(), input.Shape);
            for (int i = n_min; i < n_max; i++)
            {
                mean += images[i];
            }

            return mean / new F4(n_used);
        }
        public static F4PA ModifiedTrimmedMean(F4PA input, params object[] parameters)
        {
            var radius = (int)parameters[0];
            var alfa = new F4((float)parameters[1]);

            var zero = new F4(0);
            var one = new F4(1);

            var n_samples = (2 * radius + 1) * (2 * radius + 1);
            var n_half = n_samples / 2;

            var images = new F4PA[n_samples];

            var ctr = 0;
            for (int i = -radius; i <= radius; i++)
                for (int j = -radius; j <= radius; j++)
                    images[ctr++] = PA.Shift(input, new int[] { i, j });

            FilterCore.Sort(images);

            var median = images[n_half];
            var mean = new F4PA(new F4(), input.Shape);
            var count = new F4PA(new F4(), input.Shape);

            for (int i = 0; i < images.Length; i++)
            {
                var cond = alfa - PA.Abs(images[i] - median);// for all positive diff pixels
                mean += PA.Select(cond, images[i], zero); // sum the pixel values
                count += PA.Select(cond, one, zero); // and count number of sumations
            }
            return FilterCore.Divide(mean, count);
        }
        public static F4PA KNearestNeighbor(F4PA input, params object[] parameters)
        {
            var radius = (int)parameters[0];
            var k_int = (int)parameters[1];
            var k_f4 = new F4PA(new F4(1 + (int)parameters[1]), input.Shape);

            var zero = new F4(0);
            var one = new F4(1);

            var n_samples = (2 * radius + 1) * (2 * radius + 1);
            var n_half = n_samples / 2;

            var images = new F4PA[n_samples];// all the images
            var diff = new F4PA[n_samples];// the difference between centre element and a given image

            var ctr = 0;
            for (int i = -radius; i <= radius; i++)
                for (int j = -radius; j <= radius; j++)
                    images[ctr++] = PA.Shift(input, new int[] { i, j });

            var center = images[n_half];

            for (int i = 0; i < images.Length; i++)
                diff[i] = PA.Abs(images[i] - center);

            FilterCore.Sort(diff, images);// Sorting images with respect to diff

            var sum = new F4PA(new F4(), input.Shape);
            for (int i = 0; i <= k_int; i++)
                sum += images[i];

            return FilterCore.Divide(sum, k_f4);
        }

    }
}
