﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Microsoft.ParallelArrays;
using PA = Microsoft.ParallelArrays.ParallelArrays;
using BPA = Microsoft.ParallelArrays.BoolParallelArray;
using DPA = Microsoft.ParallelArrays.DoubleParallelArray;
using FPA = Microsoft.ParallelArrays.FloatParallelArray;
using IPA = Microsoft.ParallelArrays.IntParallelArray;
using F4PA = Microsoft.ParallelArrays.Float4ParallelArray;
using System.Windows.Media;

namespace ImageLab.Filter
{
    public static class MedianFilter
    {
        /// <summary>
        /// O(input.Size) x O(1)
        /// </summary>
        /// <param name="input"></param>
        /// <returns></returns>
        private static byte[,] MedianRowHistogram(byte[,] input, int radius)
        {
            var cols = input.GetLength(0); var rows = input.GetLength(1);

            var output = new byte[cols, rows];

            var histogram_array = new HistogramArray(rows, radius);
            var histogram_win = new MultilevelHistogram();

            var image = new ByteImage(input);
            var window = image.Window;

            window.Radius = radius;

            for (int y = 0; y < image.Height; y++)
                for (int x = 0; x < radius; x++)
                    histogram_array[y].Add(image[x, y]);

            for (int x = 0; x < image.Width; x++)
            {
                // Window histogram is cleared and made ready for new row
                histogram_win.Clear();
                for (int y = 0; y < radius; y++)
                {
                    var currentRow = histogram_array[y];
                    var remove = new Coordinate(x - radius - 1, y);
                    if (image.Contains(remove))
                    {
                        currentRow.Remove(image[remove]);
                    }
                    var add = new Coordinate(x + radius, y);
                    if (image.Contains(add))
                    {
                        currentRow.Add(image[add]);
                    }
                    histogram_win.Add(currentRow);
                }// for
                for (int y = 0; y < image.Height; y++)
                {
                    histogram_array.Position = y;

                    // Removing the histogram that is now outside the window, O(1)
                    var over = histogram_array.OutsideTop;
                    if (over != null) histogram_win.Remove(histogram_array.OutsideTop);

                    var bottom = histogram_array.WindowBottom;
                    if (bottom != null)
                    {
                        // Removing the pixel that no longer is in the window, O(1)
                        var remove = window.BottomLeft + Coordinate.Left;
                        if (image.Contains(remove))
                        {
                            bottom.Remove(image[remove]);
                        } // if

                        // Adding the pixel that just entered the window, O(1)                        
                        var add = window.BottomRight;
                        if (image.Contains(add))
                        {
                            bottom.Add(image[add]);
                        } // if
                        // Adding the histogram that is now inside the window, O(1)
                        histogram_win.Add(bottom);
                    } // if

                    // Calculating the Median, O(1)
                    output[x, y] = histogram_win.Median;

                    window.Move(Coordinate.Down);
                } // for 
                window.MoveToTop();
                window.Move(Coordinate.Right);

                //Console.WriteLine(100f * x / (float)(image.Width - 1) + "%");
            } // for

            return output;
        } // method

        public static F4PA Median(F4PA input, params object[] parameters)
        {
            var radius = (int)parameters[0];
            if (radius == 0) return input;

            var n_samples = (2 * radius + 1) * (2 * radius + 1);
            var n_half = (n_samples - 1) / 2 + 1;
            var pa_half = new F4PA(new Float4(n_half), input.Shape);

            var pa_min = new F4PA(new Float4(), input.Shape);
            var pa_max = new F4PA(new Float4(1), input.Shape);

            var pa_win = new F4PA[n_samples];
            for (int x = -radius, i = 0; x <= radius; x++)
            {
                for (int y = -radius; y <= radius; y++, i++)
                {
                    pa_win[i] = PA.Shift(input, new int[] { x, y });
                }
            }

            for (int i_run = 0; i_run < 8; i_run++)
            {
                var pa_pivot = (pa_max + pa_min) / new Float4(2);
                var pa_count = new F4PA(new Float4(), input.Shape);

                for (int i = 0; i < pa_win.Length; i++)
                    pa_count += PA.Select(pa_win[i] - pa_pivot, new Float4(1), new Float4(0));

                var pa_diff = pa_count - pa_half;

                pa_min = PA.Select(pa_diff, pa_pivot, pa_min);
                pa_max = PA.Select(-pa_diff, pa_pivot, pa_max);
            }

            return pa_min;
        }

    }
}
