﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ImageLab.Filter
{
    public class MultilevelHistogram : IHistogram
    {
        public Histogram Top { get; private set; }
        public Histogram[] Bottom { get; private set; }

        public int[] Frequency
        {
            get
            {
                var output = new int[256];

                for (int i = 0; i < 16; i++)
                {
                    var top = Top[i];
                    if (top != 0)
                    {
                        for (int j = 0; j < 16; j++)
                        {
                            var bottom = Bottom[i][j];
                            if (bottom != 0)
                            {
                                output[i * 16 + j] = bottom;
                            }
                        }
                    }
                }
                return output;
            }
        }
        public int Samples { get; private set; }

        public MultilevelHistogram()
        {
            Top = new Histogram(16);
            Bottom = new Histogram[16];
            for (int i = 0; i < Bottom.Length; i++)
            {
                Bottom[i] = new Histogram(16);
            }
        }

        public void Add(Histogram input)
        {
            for (int i = 0; i < input.Length; i++)
            {
                var observations = input[i];
                if (observations != 0)
                {
                    var cast = (int)i;
                    byte top = (byte)(cast / 16);
                    byte bottom = (byte)(cast & 0x0f);

                    Top.Add(top, observations);
                    Bottom[top].Add(bottom, observations);
                }
            }
            Samples += input.Samples;
        }
        public void Add(MultilevelHistogram input)
        {
            Top.Add(input.Top);
            for (int i = 0; i < 16; i++)
            {
                Bottom[i].Add(input.Bottom[i]);
            }
            Samples += input.Samples;
        }
        public void Add(byte input)
        {
            byte top;
            byte bottom;
            SplitByte(input, out top, out bottom);

            Top.Add(top);
            Bottom[top].Add(bottom);
            Samples++;
        }
        public void Add(byte input, int number)
        {
            byte top;
            byte bottom;
            SplitByte(input, out top, out bottom);

            Top.Add(top);
            Bottom[top].Add(bottom);
            Samples += number;
        }
        public void Remove(Histogram input)
        {
            for (int i = 0; i < input.Length; i++)
            {
                var observations = input[i];
                if (observations != 0)
                {
                    var cast = (int)i;
                    byte top = (byte)(cast / 16);
                    byte bottom = (byte)(cast & 0x0f);

                    Top.Remove(top, observations);
                    Bottom[top].Add(bottom, observations);
                }
            }
            Samples -= input.Samples;
        }
        public void Remove(byte input)
        {
            byte top;
            byte bottom;
            SplitByte(input, out top, out bottom);

            Top.Remove(top);
            Bottom[top].Remove(bottom);
            Samples--;
        }
        public void Remove(byte input, int number)
        {
            byte top;
            byte bottom;
            SplitByte(input, out top, out bottom);

            Top.Remove(top, number);
            Bottom[top].Remove(bottom, number);
            Samples -= number;
        }

        public void Clear()
        {
            Top.Clear();
            for (int i = 0; i < Bottom.Length; i++)
            {
                Bottom[i].Clear();
            }
            Samples = 0;
        }
        public byte Median
        {
            get
            {
                byte top = 0, bottom = 0;
                int sum_top = 0, sum_bottom = 0;

                for (; top < Top.Length; top++)// 16 iterations worst case, average 8
                {
                    sum_top += Top[top];
                    if (sum_top >= Samples / 2f)
                    {
                        sum_bottom = sum_top - Top[top];
                        break;
                    }
                }
                for (; bottom < Bottom[top].Length; bottom++)//16 iterations worst case, average 8
                {
                    sum_bottom += Bottom[top][bottom];
                    if (sum_bottom >= Samples / 2f) break;
                }

                return (byte)(16 * top + bottom);
            }
        }

        public override string ToString()
        {
            var freq = Frequency;
            string output = "{";

            for (int i = 0; i < freq.Length; i++)
            {
                var observed = freq[i];
                for (int j = 0; j < observed; j++)
                {
                    output += i + ", ";
                }
            }
            output += "}";
            return output;
        }

        private static void SplitByte(byte input, out byte top, out byte bottom)
        {
            var cast = (int)input;
            top = (byte)(cast / 16);
            bottom = (byte)(cast & 0x0f);
        }
    }
}
