﻿using System;
using System.Linq;
using Microsoft.ParallelArrays;

using F4PA = Microsoft.ParallelArrays.Float4ParallelArray;
using FPA = Microsoft.ParallelArrays.FloatParallelArray;
using IPA = Microsoft.ParallelArrays.IntParallelArray;
using PA = Microsoft.ParallelArrays.ParallelArrays;

namespace ImageLab.Filter
{
    public static class Noise
    {
        public static F4PA Impulsive(F4PA input, params object[] parameters)
        {
            var probability = new F4PA(
                new Float4((float)parameters[0] / 2),
                input.Shape);

            var one = new F4PA(new Float4(1), input.Shape);

            input = PA.Floor(input * new Float4(256));
            for (int i = 0; i < 8; i++)
            {
                var randomFactor1 = GenerateNoiseF4PA(input.Shape[1], input.Shape[0]);// [0,1] if larger then probability, then error will occur
                var factor = new Float4((float)Math.Pow(2, i));

                input = PA.Select(
                    probability - randomFactor1,
                    input + factor,
                    input); // up

                input = PA.Select(
                    randomFactor1 + probability - one,
                    input - factor,
                    input); // down
            }
            input /= new Float4(256);

            return input;
        }

        public static F4PA Uniform(F4PA input, params object[] parameters)
        {
            var multiplier = (float)(double)parameters[0];// Distance to samples used.

            var pa_noisef4 = GenerateNoiseF4PA(input.Shape[1], input.Shape[0]);// [0,1]
            pa_noisef4 -= new Float4(0.5f, 0.5f, 0.5f, 0);// [-.5,.5]
            pa_noisef4 *= new Float4(multiplier, multiplier, multiplier, 1);

            return input + pa_noisef4;
        }
        public static F4PA Gaussian(F4PA input, params object[] parameters)
        {
            var mean = (float)(double)parameters[0];// 
            var std = (float)(double)parameters[1];// 

            var pa_noisef4 = GenerateNoiseF4PA(input.Shape[1], input.Shape[0]);// [0,1]

            var pa_r = (PA.FromFloat4ParallelArray(pa_noisef4, 1));
            var pa_g = (PA.FromFloat4ParallelArray(pa_noisef4, 2));
            var pa_b = (PA.FromFloat4ParallelArray(pa_noisef4, 3));
            var pa_a = (new FPA(1, pa_r.Shape));

            var r = FilterCore.DX9.ToArray2D(InverseNormalCDF(pa_r));
            var g = FilterCore.DX9.ToArray2D(InverseNormalCDF(pa_g));
            var b = FilterCore.DX9.ToArray2D(InverseNormalCDF(pa_b));

            pa_r = new FPA(r);
            pa_g = new FPA(g);
            pa_b = new FPA(b);

            var pa_noise = PA.ToFloat4ParallelArray(pa_r, pa_g, pa_b, pa_a);
            pa_noise = new Float4(mean) + new Float4(std) * pa_noise;

            return input + pa_noise;
        }
        public static F4PA GenerateNoiseF4PA(int width, int height)
        {
            var noise = new int[height, width * 3];
            var buffer = new byte[3 * 2 * height * width];
            FilterCore.RNG.GetBytes(buffer);
            var buffer_short = new ushort[buffer.Length / 2];
            Buffer.BlockCopy(buffer, 0, buffer_short, 0, buffer.Length);

            var ctr = 0;
            for (int i = 0; i < noise.GetLength(0); i++)
            {
                for (int j = 0; j < noise.GetLength(1); j++, ctr++)
                {
                    noise[i, j] = buffer_short[ctr];
                }
            }

            var pa_noise = PA.ToFloatParallelArray(new IPA(noise));

            var ssr = new SectionSpecifier(0, -1, 3);// every 4th element, offset 0
            var ssg = new SectionSpecifier(1, -1, 3);// every 4th element, offset 1
            var ssb = new SectionSpecifier(2, -1, 3);// every 4th element, offset 2
            var ssother = new SectionSpecifier(0, -1, 1);// every element
            // Partitioning the image
            var r = PA.Section(pa_noise, ssother, ssr);
            var g = PA.Section(pa_noise, ssother, ssg);
            var b = PA.Section(pa_noise, ssother, ssb);
            var a = new FloatParallelArray(1, new int[] { height, width });

            var pa_noisef4 = PA.ToFloat4ParallelArray(r, g, b, a);
            pa_noisef4 /= new Float4(ushort.MaxValue, ushort.MaxValue, ushort.MaxValue, 1);//[0,1]
            pa_noisef4 *= new Float4(1, 1, 1, 0);//[0,1]
            pa_noisef4 += new Float4(0, 0, 0, 1);//[0,1]

            return pa_noisef4;
        }


        //Coefficients in rational approximations.
        private static readonly double[] a = new double[]{
            -3.969683028665376e+01,
            2.209460984245205e+02,
            -2.759285104469687e+02,
            1.383577518672690e+02,
            -3.066479806614716e+01,
            2.506628277459239e+00 };

        private static readonly double[] b = new double[]{
            -5.447609879822406e+01,
            1.615858368580409e+02,
            -1.556989798598866e+02,
            6.680131188771972e+01,
            -1.328068155288572e+01};

        private static readonly double[] c = new double[]{
            -7.784894002430293e-03,
            -3.223964580411365e-01,
            -2.400758277161838e+00,
            -2.549732539343734e+00,
            4.374664141464968e+00,
            2.938163982698783e+00};

        private static readonly double[] d = new double[] {
            7.784695709041462e-03,
            3.224671290700398e-01,
            2.445134137142996e+00,
            3.754408661907416e+00 };

        /// <summary>
        /// Algorithm described on http://home.online.no/~pjacklam/notes/invnorm/, as
        /// pseudocode, translated on 08.06.11
        /// </summary>
        /// <param name="input"></param>
        /// <returns></returns>
        public static double InverseNormalCDF(double input)
        {
            double output = 0;

            if (input == 0) output = Double.NegativeInfinity;
            else if (input == 1) output = Double.PositiveInfinity;
            else if (Double.IsNaN(input) || input < 0 || input > 1) output = Double.NaN;

            // Define break-points.
            var input_low = 0.02425;
            var input_high = 1 - input_low;

            // Rational approximation for lower region.
            if (0 < input && input < input_low)
            {
                var q = Math.Sqrt(-2 * Math.Log(input));
                output = (((((c[0] * q + c[1]) * q + c[2]) * q + c[3]) * q + c[4]) * q + c[5]) /
                      ((((d[0] * q + d[1]) * q + d[2]) * q + d[3]) * q + 1);
            }

            // Rational approximation for central region.
            if (input_low <= input && input <= input_high)
            {
                var q = input - 0.5;
                var r = q * q;
                output = (((((a[0] * r + a[1]) * r + a[2]) * r + a[3]) * r + a[4]) * r + a[5]) * q /
                     (((((b[0] * r + b[1]) * r + b[2]) * r + b[3]) * r + b[4]) * r + 1);
            }

            // Rational approximation for upper region.
            if (input_high < input && input < 1)
            {
                var q = Math.Sqrt(-2 * Math.Log(1 - input));
                output = -(((((c[0] * q + c[1]) * q + c[2]) * q + c[3]) * q + c[4]) * q + c[5]) /
                       ((((d[0] * q + d[1]) * q + d[2]) * q + d[3]) * q + 1);
            }
            return output;
        }
        /// <summary>
        /// Same algortihm as above, just modified for parallel processing.
        /// Using floats because the DX9 target cannot handle some complex
        /// operations with doubles, ie. it doesn't support square root,
        /// or logarithm operations.
        /// </summary>
        /// <param name="input"></param>
        /// <returns></returns>
        public static FPA InverseNormalCDF(FPA input)
        {
            var output = new FPA(0, input.Shape);

            var zero = new FPA(0, input.Shape);
            var one = new FPA(1, input.Shape);
            var E = new FPA((float)Math.E, input.Shape);
            var inf_p = new FPA(float.PositiveInfinity, input.Shape);
            var inf_n = new FPA(float.NegativeInfinity, input.Shape);
            var nan = new FPA(float.NaN, input.Shape);

            var input_low = new FPA(0.02425f, input.Shape);
            var input_high = new FPA(1 - 0.02425f, input.Shape);
            /*
            output = PA.Cond(PA.CompareEqual(input, zero), inf_n, output);
            output = PA.Cond(PA.CompareEqual(input, one), inf_p, output);            
            output = PA.Cond(
                PA.CompareEqual(input, one) | input < zero | input > one,
                nan, output);
            */
            var a = (from x in Noise.a select (float)x).ToArray();
            var b = (from x in Noise.b select (float)x).ToArray();
            var c = (from x in Noise.c select (float)x).ToArray();
            var d = (from x in Noise.d select (float)x).ToArray();

            // Rational approximation for lower region.
            var q_l1 = PA.Sqrt(-2.0f * PA.Log2(input) / PA.Log2(E));
            var q_l2 = (((((c[0] * q_l1 + c[1]) * q_l1 + c[2]) * q_l1 + c[3]) * q_l1 + c[4]) * q_l1 + c[5]) /
                      ((((d[0] * q_l1 + d[1]) * q_l1 + d[2]) * q_l1 + d[3]) * q_l1 + 1);

            var cond_l = zero < input & input < input_low;
            output = PA.Cond(cond_l, q_l2, output);

            // Rational approximation for central region.
            var q_c1 = input - 0.5f;
            var r_c = q_c1 * q_c1;
            var q_c2 = (((((a[0] * r_c + a[1]) * r_c + a[2]) * r_c + a[3]) * r_c + a[4]) * r_c + a[5]) * q_c1 /
                     (((((b[0] * r_c + b[1]) * r_c + b[2]) * r_c + b[3]) * r_c + b[4]) * r_c + 1);

            var cond_c = input_low <= input & input <= input_high;
            output = PA.Cond(cond_c, q_c2, output);

            // Rational approximation for upper region.
            var q_u1 = PA.Sqrt(-2 * PA.Log2(1 - input) / PA.Log2(E));
            var q_u2 = -(((((c[0] * q_u1 + c[1]) * q_u1 + c[2]) * q_u1 + c[3]) * q_u1 + c[4]) * q_u1 + c[5]) /
                       ((((d[0] * q_u1 + d[1]) * q_u1 + d[2]) * q_u1 + d[3]) * q_u1 + 1);

            var cond_u = input_high < input & input < one;
            output = PA.Cond(cond_u, q_u2, output);

            return output;
        }
    }
}
