﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;

using ImageLab.Filter;

namespace ImageLab
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private WriteableBitmap bitmap;
        private List<WriteableBitmap> undoChain;
        private string currentImagePath;
        private IDictionary<string, Window> openWindows;
        private IDictionary<string, FilterCore.FilterDelegate> filters;
        private Window histogramWindow;
        private Graph histogramGraph;

        private IDictionary<string, object[]> parameters;

        public MainWindow()
        {
            InitializeComponent();
        }
        protected override void OnInitialized(EventArgs e)
        {
            this.bitmap = new WriteableBitmap((BitmapSource)imageControl.Source);
            ConvertImageFormat();
            this.imageControl.Source = this.bitmap;

            undoChain = new List<WriteableBitmap>();

            openWindows = new Dictionary<string, Window>();
            filters = new Dictionary<string, FilterCore.FilterDelegate>();
            parameters = new Dictionary<string, object[]>();

            //TODO: Move this into separate classes for registering

            filters["Mean Filter"] = MeanFilter.Mean;
            filters["Median Filter"] = MedianFilter.Median;
            filters["Distance Weighted Median Filter"] = DistanceWeightedMedian.DWMedian;
            filters["Hybrid Mean-Median Filter"] = HybridMeanMedianFilter.HybridMeanMedian;
            filters["Hybrid Mean-Distance Weighted Median Filter"] = HybridMeanMedianFilter.HybridMeanDWMedian;
            filters["Winsorized Mean Filter"] = MeanFilter.WinsorizedMean;
            filters["Modified Trimmed Mean Filter"] = MeanFilter.ModifiedTrimmedMean;
            filters["K-Nearest Neighbor Filter"] = MeanFilter.KNearestNeighbor;

            filters["Uniform Noise"] = Noise.Uniform;
            filters["Gaussian Noise"] = Noise.Gaussian;
            filters["Impulsive Noise"] = Noise.Impulsive;

            filters["Stretch"] = Transforms.Stretch;
            filters["Transpose"] = Transforms.Transpose;
            filters["Brightness/Contrast"] = Transforms.BrightnessContrast;
            filters["Gamma Correction"] = Transforms.GammaCorrection;
            filters["Black and White"] = Transforms.ToBlackWhite;

            filters["Variance"] = Analysis.Variance;

            parameters["Mean Filter"] = new object[1];
            parameters["Median Filter"] = new object[1];
            parameters["Distance Weighted Median Filter"] = new object[1];
            parameters["Hybrid Mean-Median Filter"] = new object[2];
            parameters["Hybrid Mean-Distance Weighted Median Filter"] = new object[2];
            parameters["Winsorized Mean Filter"] = new object[2];
            parameters["Modified Trimmed Mean Filter"] = new object[2];
            parameters["K-Nearest Neighbor Filter"] = new object[2];

            parameters["Uniform Noise"] = new object[1];
            parameters["Gaussian Noise"] = new object[2];
            parameters["Impulsive Noise"] = new object[1];

            parameters["Stretch"] = new object[2];
            parameters["Transpose"] = new object[0];
            parameters["Brightness/Contrast"] = new object[2];
            parameters["Gamma Correction"] = new object[1];
            parameters["Black and White"] = new object[1];

            parameters["Variance"] = new object[1];

            base.OnInitialized(e);
        }

        public WriteableBitmap Bitmap
        {
            get { return bitmap; }
            set { bitmap = value; }
        }
        /// <summary>
        /// Will allways allow opening
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void OpenCommandBinding_CanExecute(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = true;
        }

        private void SaveCommandBinding_CanExecute(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = true;
        }

        private void SaveAsCommandBinding_CanExecute(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = true;
        }
        private void CloseCommandBinding_CanExecute(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = true;
        }
        private void OpenCommandBinding_Executed(object sender, ExecutedRoutedEventArgs e)
        {
            Microsoft.Win32.OpenFileDialog dlg = new Microsoft.Win32.OpenFileDialog();
            dlg.FileName = "Image"; // Default file name
            dlg.DefaultExt = ".jpg"; // Default file extension
            dlg.Filter = "Images |*.bmp;*.gif;*.ico;*.jpg;*.png;*.wdp;*.tiff|All Files|*.*"; // Filter files by extension

            // Show open file dialog box
            Nullable<bool> result = dlg.ShowDialog();

            // Process open file dialog box results
            if (result == true)
            {
                // Open document
                currentImagePath = dlg.FileName;
                OpenImage(currentImagePath);
            }
        }
        private void OpenImage(string filename)
        {
            bitmap = new WriteableBitmap((BitmapSource)new BitmapImage(new Uri(filename)));
            ConvertImageFormat();
            imageControl.Source = bitmap;
            UpdateHistogram();
        }
        private void SaveCommandBinding_Executed(object sender, ExecutedRoutedEventArgs e)
        {
            if (currentImagePath == null)
            {
                SaveAsCommandBinding_Executed(sender, e);
            }
            else
            {
                SaveImage(currentImagePath);
            }
        }
        private void SaveAsCommandBinding_Executed(object sender, ExecutedRoutedEventArgs e)
        {
            // Configure save file dialog box            
            Microsoft.Win32.SaveFileDialog dlg = new Microsoft.Win32.SaveFileDialog();
            dlg.FileName = "Image"; // Default file name
            dlg.DefaultExt = ".jpg"; // Default file extension
            dlg.Filter = "JPeg Image(*.JPG)|*.jpg|Bitmap Image(*.BMP)|*.bmp|Png Image(*.PNG)|*.png|Gif Image(*.GIF)|*.gif";

            // Show save file dialog box
            Nullable<bool> result = dlg.ShowDialog();

            // Process save file dialog box results
            if (result == true)
            {
                SaveImage(dlg.FileName);
            }
        }

        private void SaveImage(string fileName)
        {
            BitmapEncoder encoder = new BmpBitmapEncoder();

            string extension = fileName.Substring(fileName.LastIndexOf('.'));
            switch (extension.ToLower())
            {
                case ".jpg":
                    encoder = new JpegBitmapEncoder();
                    break;
                case ".bmp":
                    encoder = new BmpBitmapEncoder();
                    break;
                case ".gif":
                    encoder = new GifBitmapEncoder();
                    break;
                case ".png":
                    encoder = new PngBitmapEncoder();
                    break;
            }
            // push the rendered bitmap to it
            //encoder.Frames.Add(BitmapFrame.Create(renderBitmap));
            encoder.Frames.Add(BitmapFrame.Create(bitmap));
            // Create a file stream for saving image
            using (FileStream fs = File.Open(fileName, FileMode.OpenOrCreate))
            {
                encoder.Save(fs);
                fs.Close();
            }
        }
        private void CloseCommandBinding_Executed(object sender, ExecutedRoutedEventArgs e)
        {
            this.Close();
        }

        private void FilterMenuItem_Click(object sender, RoutedEventArgs e)
        {
            var instruction = sender as MenuItem;
            if (instruction == null) return;

            string insStr = (string)instruction.Header;
            if (!openWindows.ContainsKey(insStr))
            {
                #region New window initialization
                // Initialization
                var win = new Window();
                var wrapPanel = new WrapPanel();
                var button = new Button();
                #endregion

                // Setup
                button.Content = "Run";
                win.Title = insStr;
                win.SizeToContent = SizeToContent.WidthAndHeight;
                win.WindowStyle = System.Windows.WindowStyle.ToolWindow;
                //                wrapPanel.MaxWidth = 120;

                switch (insStr)
                {
                    case "Mean Filter":
                    case "Median Filter":
                    case "Distance Weighted Median Filter":
                    case "Variance":
                        Window_Radius(wrapPanel, insStr);
                        break;
                    case "Uniform Noise":
                        Window_1Double(wrapPanel, insStr);
                        break;
                    case "Gaussian Noise":
                        Window_2Double(wrapPanel, insStr);
                        break;
                    case "Stretch":
                        Window_2Int(wrapPanel, insStr);
                        break;
                    case "Transpose":
                        break;
                    case "Brightness/Contrast":
                        Window_BrightnessContrast(wrapPanel, insStr);
                        break;
                    case "Gamma Correction":
                        Window_Gamma(wrapPanel);
                        break;
                    case "Hybrid Mean-Distance Weighted Median Filter":
                    case "Hybrid Mean-Median Filter":
                        Window_RadiusCrossover(wrapPanel, insStr);
                        break;
                    case "Black and White":
                        Window_Treshold(wrapPanel, insStr);
                        break;
                    case "Impulsive Noise":
                        Window_Probability(wrapPanel, insStr);
                        break;
                    case "Winsorized Mean Filter":
                    case "Modified Trimmed Mean Filter":
                        Window_RadiusAlfa(wrapPanel, insStr);
                        break;
                    case "K-Nearest Neighbor Filter":
                        Window_RadiusK(wrapPanel, insStr);
                        break;
                }

                // Event handlers                                      
                #region Event Handler Setup
                button.Click += delegate(Object o, RoutedEventArgs arg)
                {
                    ConvertImageFormat();

                    undoChain.Add(bitmap);
                    FilterCore.Input = bitmap;
                    try
                    {
                        FilterCore.Process(filters[insStr], parameters[insStr]);
                        bitmap = FilterCore.Output;
                        imageControl.Source = bitmap;
                        if (histogramWindow != null)
                        {
                            histogramGraph.Data = FilterCore.Histogram;
                            histogramGraph.Update();
                        }
                    }
                    catch (Exception exc)
                    {
                        MessageBox.Show(exc.Message + "\n" + exc.StackTrace);
                        Undo();
                    }
                };
                win.Closed += delegate
                {
                    openWindows.Remove(insStr);
                };
                #endregion
                // Lay out
                #region Lay out
                wrapPanel.Children.Add(button);
                win.Content = wrapPanel;
                #endregion

                openWindows[insStr] = win;
                win.Show();
            }
            else openWindows[insStr].Focus();
        }
        private void Window_RadiusK(WrapPanel wrapPanel, string insStr)
        {
            var label1 = new Label();
            var label2 = new Label();
            var slider1 = new Slider();
            var slider2 = new Slider();

            label1.Content = "Radius x1";
            label2.Content = "K x1";

            slider1.TickFrequency = 1;
            slider1.Minimum = 1;
            slider1.Value = 1;
            slider1.Maximum = 5;
            slider1.Width = 100;

            slider2.TickFrequency = 1;
            slider2.Minimum = 0;
            slider2.Value = 1;
            slider2.Maximum = 8;
            slider2.Width = 100;

            wrapPanel.Children.Add(label1);
            wrapPanel.Children.Add(slider1);
            wrapPanel.Children.Add(label2);
            wrapPanel.Children.Add(slider2);

            parameters[insStr][0] = 1;
            parameters[insStr][1] = 1;
            slider1.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                var r = (int)arg.NewValue;
                var n_max = (2 * r + 1) * (2 * r + 1) - 1;
                slider2.Maximum = n_max;
                parameters[insStr][0] = r;
                label1.Content = string.Format("Radius x{0}", (int)arg.NewValue);

            };
            slider2.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                var r = (int)parameters[insStr][0];
                var n_max = (2 * r + 1) * (2 * r + 1) - 1;
                var parameter = Math.Min((int)arg.NewValue, n_max);
                parameters[insStr][1] = parameter;
                label2.Content = string.Format("K x{0}", parameter);
            };
        }
        private void Window_RadiusAlfa(WrapPanel wrapPanel, string insStr)
        {
            var label_radius = new Label();
            var label_alfa = new Label();
            var slider_radius = new Slider();
            var slider_alfa = new Slider();

            label_radius.Content = "Radius x1";
            slider_radius.IsSnapToTickEnabled = true;
            slider_radius.TickFrequency = 1;
            slider_radius.Minimum = 0;
            slider_radius.Maximum = 16;
            slider_radius.Value = 1;
            slider_radius.Width = 80;

            label_alfa.Content = "Alfa x0.10";
            slider_alfa.Minimum = 0;
            slider_alfa.Maximum = 0.5;
            slider_alfa.Value = 0.1;
            slider_alfa.Width = 80;

            wrapPanel.Children.Add(label_radius);
            wrapPanel.Children.Add(slider_radius);
            wrapPanel.Children.Add(label_alfa);
            wrapPanel.Children.Add(slider_alfa);

            parameters[insStr][0] = 1;
            parameters[insStr][1] = 0.1f;
            slider_radius.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                parameters[insStr][0] = (int)arg.NewValue;
                label_radius.Content = string.Format("Radius x{0}", (int)arg.NewValue);

            };
            slider_alfa.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                parameters[insStr][1] = (float)arg.NewValue;
                label_alfa.Content = string.Format("Alfa x{0:0.00}", (float)arg.NewValue);
            };
        }
        private void Window_Probability(WrapPanel wrapPanel, string insStr)
        {
            var label = new Label();
            var slider = new Slider();

            label.Content = "Probability x0.50";
            slider.Minimum = 0;
            slider.Maximum = 1;
            slider.TickFrequency = .01;
            slider.Value = 0.5;
            slider.Width = 100;

            wrapPanel.Children.Add(label);
            wrapPanel.Children.Add(slider);

            parameters[insStr][0] = 0.5f;

            slider.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                parameters[insStr][0] = (float)arg.NewValue;
                label.Content = string.Format("Treshold x{0:0.00}", arg.NewValue);
            };
        }
        private void Window_Treshold(WrapPanel wrapPanel, string insStr)
        {
            var label = new Label();
            var slider = new Slider();

            label.Content = "Treshold x0.50";
            slider.Minimum = 0;
            slider.Maximum = 1;
            slider.TickFrequency = .01;
            slider.Value = 0.5;
            slider.Width = 100;

            wrapPanel.Children.Add(label);
            wrapPanel.Children.Add(slider);

            parameters[insStr][0] = 0.5f;

            slider.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                parameters[insStr][0] = (float)arg.NewValue;
                label.Content = string.Format("Treshold x{0:0.00}", arg.NewValue);
            };
        }

        private void Window_RadiusCrossover(WrapPanel wrapPanel, string insStr)
        {
            var label_radius = new Label();
            var label_crossover = new Label();
            var slider_radius = new Slider();
            var slider_crossover = new Slider();

            label_radius.Content = "Radius x1";
            slider_radius.IsSnapToTickEnabled = true;
            slider_radius.TickFrequency = 1;
            slider_radius.Minimum = 0;
            slider_radius.Maximum = 16;
            slider_radius.Value = 1;
            slider_radius.Width = 80;

            label_crossover.Content = "Crossover x0.01";
            slider_crossover.Minimum = 0;
            slider_crossover.Maximum = 0.2;
            slider_crossover.Value = 0.1;
            slider_crossover.Width = 80;

            wrapPanel.Children.Add(label_radius);
            wrapPanel.Children.Add(slider_radius);
            wrapPanel.Children.Add(label_crossover);
            wrapPanel.Children.Add(slider_crossover);

            parameters[insStr][0] = 1;
            parameters[insStr][1] = 0.1f;
            slider_radius.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                parameters[insStr][0] = (int)arg.NewValue;
                label_radius.Content = string.Format("Radius x{0}", (int)arg.NewValue);

            };
            slider_crossover.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                parameters[insStr][1] = (float)arg.NewValue;
                label_crossover.Content = string.Format("Crossover x{0:0.00}", (float)arg.NewValue);
            };
        }
        private void Window_Gamma(WrapPanel wrapPanel)
        {
            var label = new Label();
            var slider = new Slider();

            label.Content = "Gamma x1.0";
            slider.Minimum = 0;
            slider.Maximum = 3;
            slider.Value = 1;
            slider.TickFrequency = .01;
            slider.Width = 100;

            wrapPanel.Children.Add(label);
            wrapPanel.Children.Add(slider);

            parameters["Gamma Correction"][0] = 1f;

            slider.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                parameters["Gamma Correction"][0] = (float)arg.NewValue;
                label.Content = string.Format("Gamma x{0:0.0}", arg.NewValue);
            };
        }
        private void Window_BrightnessContrast(WrapPanel wrapPanel, string insStr)
        {
            var label1 = new Label();
            var label2 = new Label();
            var slider1 = new Slider();
            var slider2 = new Slider();

            slider1.Minimum = -1;
            slider1.Value = 0;
            slider1.Maximum = 1;
            slider1.Width = 100;
            slider1.TickFrequency = 0.01;
            slider1.IsSnapToTickEnabled = true;
            slider1.SmallChange = 0.01;
            slider1.LargeChange = 0.1;
            label1.Content = string.Format("Brightness x{0:+0.00;-0.00;+0.00}", slider1.Value);

            slider2.Minimum = 0.0;
            slider2.Value = 1;
            slider2.Maximum = 2;
            slider2.Width = 100;
            slider2.TickFrequency = 0.01;
            slider2.IsSnapToTickEnabled = true;
            slider2.SmallChange = 0.01;
            slider2.LargeChange = 0.1;
            label2.Content = string.Format("Contrast x{0:0.00}", slider2.Value);


            wrapPanel.Children.Add(label1);
            wrapPanel.Children.Add(slider1);
            wrapPanel.Children.Add(label2);
            wrapPanel.Children.Add(slider2);

            parameters[insStr][0] = (float)slider1.Value;
            parameters[insStr][1] = (float)slider2.Value;

            slider1.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                parameters[insStr][0] = (float)arg.NewValue;
                label1.Content = string.Format("Brightness x{0:+0.00;-0.00;+0.00}", slider1.Value);

            };
            slider2.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                parameters[insStr][1] = (float)arg.NewValue;
                label2.Content = string.Format("Contrast x{0:0.00}", slider2.Value);
            };
        }
        private void Window_Radius(WrapPanel wrapPanel, string insStr)
        {
            var label = new Label();
            var slider = new Slider();

            label.Content = "Radius x1";
            slider.IsSnapToTickEnabled = true;
            slider.TickFrequency = 1;
            slider.Minimum = 0;
            slider.Maximum = 16;
            slider.Value = 1;
            slider.Width = 100;

            wrapPanel.Children.Add(label);
            wrapPanel.Children.Add(slider);

            parameters[insStr][0] = 1;
            slider.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                parameters[insStr][0] = (int)arg.NewValue;
                label.Content = string.Format("Radius x{0}", (int)arg.NewValue);

            };
        }
        private void Window_2Int(WrapPanel wrapPanel, string insStr)
        {
            var label1 = new Label();
            var label2 = new Label();
            var slider1 = new Slider();
            var slider2 = new Slider();

            label1.Content = "Width x1";
            label2.Content = "Height x1";

            slider1.IsSnapToTickEnabled = true;
            slider1.TickFrequency = 1;
            slider1.Minimum = 1;
            slider1.Value = 1;
            slider1.Maximum = 5;
            slider1.Width = 100;

            slider2.IsSnapToTickEnabled = true;
            slider2.TickFrequency = 1;
            slider2.Minimum = 1;
            slider2.Value = 1;
            slider2.Maximum = 5;
            slider2.Width = 100;

            wrapPanel.Children.Add(label1);
            wrapPanel.Children.Add(slider1);
            wrapPanel.Children.Add(label2);
            wrapPanel.Children.Add(slider2);

            parameters[insStr][0] = 1;
            parameters[insStr][1] = 1;
            slider1.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                parameters[insStr][0] = (int)arg.NewValue;
                label1.Content = string.Format("Width x{0}", (int)arg.NewValue);

            };
            slider2.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                parameters[insStr][1] = (int)arg.NewValue;
                label2.Content = string.Format("Height x{0}", (int)arg.NewValue);

            };
        }
        private void Window_1Double(WrapPanel wrapPanel, string insStr)
        {
            var label = new Label();
            var slider = new Slider();

            label.Content = "Multiplier";
            slider.Minimum = 0;
            slider.Maximum = 1;
            slider.TickFrequency = .01;
            slider.Value = 0.5;
            slider.Width = 100;

            wrapPanel.Children.Add(label);
            wrapPanel.Children.Add(slider);

            parameters[insStr][0] = 0.1;

            slider.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                parameters[insStr][0] = arg.NewValue;
                label.Content = string.Format("Multiplier x{0:0.0}", arg.NewValue);
            };
        }
        private void Window_2Double(WrapPanel wrapPanel, string insStr)
        {
            var label1 = new Label();
            var label2 = new Label();
            var slider1 = new Slider();
            var slider2 = new Slider();

            slider1.Minimum = -1;
            slider1.Value = 0;
            slider1.Maximum = 1;
            slider1.Width = 100;
            slider1.TickFrequency = 0.01;
            slider1.IsSnapToTickEnabled = true;
            slider1.SmallChange = 0.01;
            slider1.LargeChange = 0.1;
            label1.Content = string.Format("Mean x{0:+0.00;-0.00;+0.00}", slider1.Value);

            slider2.Minimum = 0.0;
            slider2.Value = 1;
            slider2.Maximum = 2;
            slider2.Width = 100;
            slider2.TickFrequency = 0.01;
            slider2.IsSnapToTickEnabled = true;
            slider2.SmallChange = 0.01;
            slider2.LargeChange = 0.1;
            label2.Content = string.Format("Std x{0:0.00}", slider2.Value);


            wrapPanel.Children.Add(label1);
            wrapPanel.Children.Add(slider1);
            wrapPanel.Children.Add(label2);
            wrapPanel.Children.Add(slider2);

            parameters[insStr][0] = slider1.Value;
            parameters[insStr][1] = slider2.Value;

            slider1.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                parameters[insStr][0] = arg.NewValue;
                label1.Content = string.Format("Mean x{0:+0.00;-0.00;+0.00}", slider1.Value);

            };
            slider2.ValueChanged += delegate(Object o, RoutedPropertyChangedEventArgs<double> arg)
            {
                parameters[insStr][1] = arg.NewValue;
                label2.Content = string.Format("Std x{0:0.00}", slider2.Value);
            };
        }
        private void ConvertImageFormat()
        {
            // if the bitmap is not the required format, convert it
            if (bitmap.Format != PixelFormats.Rgba128Float)
            {
                var pro = Profile.Start("Converting from " + bitmap.Format + " to " + PixelFormats.Rgba128Float);
                var fcb = new FormatConvertedBitmap(
                    bitmap,
                    PixelFormats.Rgba128Float,
                    null, 0);
                bitmap = new WriteableBitmap(fcb);
                pro.Stop();
            }
        }
        private void Window_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            try
            {
                foreach (var window in openWindows.Values)
                    window.Close();
            }
            catch (Exception) { }
        }

        private void Undo_CanExecute(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = undoChain.Count > 0;
        }
        private void Undo_Executed(object sender, ExecutedRoutedEventArgs e)
        {
            Undo();
        }
        private void Undo()
        {
            var N = undoChain.Count;
            if (N > 0)
            {
                bitmap = undoChain[N - 1];
                undoChain.RemoveAt(N - 1);
                imageControl.Source = bitmap;
                UpdateHistogram();
            }
        }

        private void HistogramMenuItem_Click(object sender, RoutedEventArgs e)
        {
            if (histogramWindow != null) return;

            histogramWindow = new Window();
            histogramGraph = new Graph();

            histogramWindow.Content = histogramGraph;
            histogramWindow.Title = "Histogram";
            histogramWindow.SizeToContent = SizeToContent.WidthAndHeight;
            histogramWindow.WindowStyle = System.Windows.WindowStyle.ToolWindow;

            histogramWindow.Show();

            histogramWindow.Closed += delegate
            {
                FilterCore.CalculateHistogram = false;
                histogramWindow = null;
                histogramGraph = null;
            };

            // The first time the image will be run to a null filter to calculate the histogram
            UpdateHistogram();
        }

        private void UpdateHistogram()
        {
            if (histogramWindow ==  null) return;
            ConvertImageFormat();
            FilterCore.CalculateHistogram = true;
            FilterCore.Input = bitmap;
            FilterCore.Process(Analysis.Null, new object[0]);
            histogramGraph.Data = FilterCore.Histogram;
            histogramGraph.Update();
        }



        private void imageControl_MouseMove(object sender, MouseEventArgs e)
        {
            var pos = e.MouseDevice.GetPosition(imageControl);
            var x = (int)(pos.X * bitmap.PixelWidth / imageControl.ActualWidth);
            var y = (int)(pos.Y * bitmap.PixelHeight / imageControl.ActualHeight);

            var buffer = new byte[16];
            bitmap.CopyPixels(
                new Int32Rect(x % bitmap.PixelWidth, y % bitmap.PixelHeight, 1, 1),
                buffer,
                bitmap.BackBufferStride,
                0);

            var r = BitConverter.ToSingle(buffer, 0);
            var g = BitConverter.ToSingle(buffer, 4);
            var b = BitConverter.ToSingle(buffer, 8);
            var a = BitConverter.ToSingle(buffer, 12);

            imageStatusText.Text = string.Format(
                "({0},{1}) R:{2:0.000}, G:{3:0.000}, B:{4:0.000}, A:{5:0.000}",
                x, y, r, g, b, a);

        }

    }
}
